﻿using UnityEngine;
using System.Collections.Generic;

//a destructible is something that can take damage and be destroyed.
//we can subclass this to different sorts of behaviors when we are
//destroyed or take damage, like play particles or animations.
//destructibles can also be healed, up to their maximum hit points.
public class Destructible : MonoBehaviour
{
    [Tooltip("The maximum hit points we can have, also how many we start with.")]
    [SerializeField]
    protected float maximumHitPoints = 100.0f;

    [Tooltip("If the descructible should destory the gameobject on death")]
    [SerializeField]
    protected bool destroySelf = false;

    protected List<DestructibleModifier> destructibleModifiers;
    

    public virtual float HitPoints
    {
        get;
        protected set;
    }
    
    public virtual bool IsDying
    {
        get;
        protected set;
    }

    public virtual bool IsInvulnerable
    {
        get;
        set;
    }
    
    public virtual void Awake() 
    {
        destructibleModifiers = new List<DestructibleModifier>( GetComponentsInChildren<DestructibleModifier>() );
    }

    public virtual void Start()
    {
        HitPoints = maximumHitPoints;
    }
    
    public virtual void TakeDamage( float amount, Destructor source = null )
    {
        ModifyHitPoints( -amount, source );
    }

    public virtual void RecoverHitPoints( float amount, Destructor source = null  )
    {
        ModifyHitPoints( amount, source );
    }
    
    public virtual void ModifyHitPoints( float amount, Destructor source = null  )
    {
        if ( IsInvulnerable && amount < 0 )
        {
            return;
        }
        
        for ( int modifierIndex = 0; modifierIndex < destructibleModifiers.Count; modifierIndex++ )
        {
            amount = destructibleModifiers[ modifierIndex ].ModifyHitPointsChange( amount, source );
        }

        if ( amount == 0.0f )
        {
            return;
        }

        HitPoints += amount;
        HitPoints = Mathf.Min( HitPoints, maximumHitPoints );
        bool willDie = HitPoints <= 0.0f;

        for ( int modifierIndex = 0; modifierIndex < destructibleModifiers.Count; modifierIndex++ )
        {
            destructibleModifiers[ modifierIndex ].HitPointsChanged( amount, source, willDie );
        }

        if ( willDie )
        {
            Die( source );
        }
    }

    public virtual void Die( Destructor source = null )
    {
        if ( IsDying )
        {
            return;
        }

        bool shouldDie = true;
        for ( int modifierIndex = 0; modifierIndex < destructibleModifiers.Count; modifierIndex++ )
        {
            shouldDie = destructibleModifiers[ modifierIndex ].ModifyShouldDie( shouldDie, source );
        }
        
        if ( shouldDie )
        {
            IsDying = true;

            for ( int modifierIndex = 0; modifierIndex < destructibleModifiers.Count; modifierIndex++ )
            {
                destructibleModifiers[ modifierIndex ].Died( source );
            }

            if (destroySelf)
            {
                Object.Destroy(gameObject);
            }
        }
    }
}