﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public class World : MonoBehaviour, SceneLoader.Listener
{
    [Tooltip("These level names must match the scene names exactly. Each scene is expected to have a Level in it.")]
    [SerializeField]
    protected string[] levelNames;

    [Tooltip("This player gets instantiated in each level from this prefab.")]
    [SerializeField]
    protected PlayerController playerPrefab;

    [Tooltip("This knight gets instantiated in each level from this prefab.")]
    [SerializeField]
    protected KnightController knightPrefab;

    [Tooltip("Background music.")]
    [SerializeField]
    protected AudioClip backgroundMusic;

    [Tooltip("Play this sound when you win.")]
    [SerializeField]
    protected AudioClip[] winSounds;

    [Tooltip("Play this sound when you lose.")]
    [SerializeField]
    protected AudioClip[] loseSounds;

    protected int loadingLevelIndex;

    public Level CurrentLevel
    {
        get;
        protected set;
    }

    public int CurrentLevelIndex
    {
        get;
        protected set;
    }

    public int LevelCount
    {
        get
        {
            return levelNames.Length;
        }
    }

    public AudioClip[] WinSounds
    {
        get
        {
            return winSounds;
        }
    }

    public AudioClip[] LoseSounds
    {
        get
        {
            return loseSounds;
        }
    }

    public const string MAXIMUM_UNLOCKED_LEVEL_KEY = "maximumUnlockedLevel";
    public const string LEVEL_TO_LOAD_KEY = "levelToLoad";

    public static World Instance
    {
        get;
        private set;
    }

    public void Awake()
    {
        if ( Instance != null && Instance != this )
        {
            Destroy( gameObject );
            return;
        }

        Instance = this;
        DontDestroyOnLoad( gameObject );
    }

    public static int MaximumUnlockedLevel
    {
        get
        {
            int highestLevel = PlayerPrefs.GetInt(MAXIMUM_UNLOCKED_LEVEL_KEY, 0);
            //TODO - Hardoced hack to get final level to load!
            if( highestLevel >= 19 )
            {
                highestLevel = 21;
            }
            return highestLevel;
        }
        set
        {
            PlayerPrefs.SetInt( MAXIMUM_UNLOCKED_LEVEL_KEY, value );
        }
    }

    public static int LevelToLoad
    {
        get
        {
            return PlayerPrefs.GetInt( LEVEL_TO_LOAD_KEY, 0 );
        }
        set
        {
            PlayerPrefs.SetInt( LEVEL_TO_LOAD_KEY, value );
        }
    }

    public void Start()
    {
        SceneLoader.Instance.AddListener( this );
        LoadFirstLevel();
    }

    public void LoadFirstLevel()
    {
        MusicController.Instance.Play( backgroundMusic );
        loadingLevelIndex = -1;
        CurrentLevelIndex = LevelToLoad;
        RestartCurrentLevel();
    }

    public void GoToNextLevel()
    {
        if ( CurrentLevelIndex >= levelNames.Length - 1 )
        {
            SceneLoader.LoadScene( SceneNames.WON_GAME_SCENE );
        }
        else
        {
            int levelIndex = Mathf.Min( levelNames.Length - 1 , CurrentLevelIndex + 1 );
            MaximumUnlockedLevel = Mathf.Max( MaximumUnlockedLevel, levelIndex );
            
            LoadLevel( levelIndex );
        }
    }

    public void RestartCurrentLevel()
    {
        LoadLevel( CurrentLevelIndex );
    }

    public void LoadLevel( int levelIndex )
    {
        if ( loadingLevelIndex >= 0 )
        {
            return;
        }

        if ( levelIndex >= levelNames.Length )
        {
            SceneLoader.LoadScene( SceneNames.WON_GAME_SCENE );
            return;
        }

        loadingLevelIndex = levelIndex;
        SceneLoader.LoadScene( levelNames[ levelIndex ] );
        CurrentLevelIndex = levelIndex;
    }

    public void LoadingFadedOut()
    {
        // CurrentLevel = GameObject.FindObjectOfType<Level>();
        // PlayerController player = Instantiate<PlayerController>( playerPrefab );
        // KnightController knight = Instantiate<KnightController>( knightPrefab );

        // CurrentLevel.Initialize( player, knight );
    }

    public void LoadingCompleted()
    {
        CurrentLevel = GameObject.FindObjectOfType<Level>();

        if ( CurrentLevel == null )
        {
            Debug.LogWarning( "The World loaded something with no Level in it." );
            return;
        }

        PlayerController player = Instantiate<PlayerController>( playerPrefab );
        KnightController knight = Instantiate<KnightController>( knightPrefab );

        CurrentLevel.Initialize( player, knight );
    }

    public void LoadingFadedIn()
    {
        loadingLevelIndex = -1;
    }
}
