﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public class Mover : MonoBehaviour
{
    [Tooltip("How quickly this Mover travels.")]
    [SerializeField]
    protected float initialSpeed = 3.0f;

    public virtual Vector2 Velocity
    {
        get;
        protected set;
    }

    public virtual Vector2 Position
    {
        get
        {
            return (Vector2) transform.position;
        }
        set
        {
            transform.position = new Vector3( value.x, value.y, transform.position.z );
        }
    }

    public virtual float InitialSpeed
    {
        get
        {
            return initialSpeed;
        }
    }

    public virtual float FacingAngleDegrees
    {
        get
        {
            return FacingAngleRadians * Mathf.Rad2Deg;
        }
    }

    public virtual float FacingAngleRadians
    {
        get
        {
            return Mathf.Atan2( Velocity.y, Velocity.x );
        }
    }

    protected List<MoverModifier> modifiers;

    public virtual void Awake()
    {
        modifiers = new List<MoverModifier>( GetComponents<MoverModifier>() );
    }

    public virtual void Start()
    {
        if ( !Mathf.Approximately( initialSpeed, 0.0f ) )
        {
            Velocity = (Vector2) transform.right * initialSpeed;
        }
    }

    public virtual void Update()
    {
        UpdatePosition();
    }

    public virtual void SetVelocity( Vector2 vel )
    {
        Velocity = vel;
    }

    public virtual void ModifyVelocity( Vector2 velChange )
    {
        Velocity += velChange;
    }

    protected virtual void UpdatePosition()
    {
        foreach( MoverModifier modifier in modifiers )
        {
            if ( modifier && modifier.Enabled )
            {
                Velocity = modifier.ModifyVelocity( Velocity );
            }
        }

        Vector2 delta = Velocity * Time.deltaTime;

        foreach( MoverModifier modifier in modifiers )
        {
            if ( modifier && modifier.Enabled )
            {
                delta = modifier.ModifyFrameDelta( delta );
            }
        }

        Vector2 newPosition = (Vector2) transform.position + delta;

        foreach( MoverModifier modifier in modifiers )
        {
            if ( modifier && modifier.Enabled )
            {
                newPosition = modifier.ModifyFinalPosition( newPosition );
            }
        }

        Position = newPosition;
    }
}
