﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public class CharacterView : MonoBehaviour
{
    [Tooltip("We use this Animator to control our view.")]
    [SerializeField]
    protected Animator animator;

    [Tooltip("The main sprite renderer for the Knight.")]
    [SerializeField]
    protected SpriteRenderer mainSpriteRenderer;

    [Tooltip("If we are moving at a speed slower than this threshold, we've stopped.")]
    [SerializeField]
    protected float walkingSpeedThreshold = 0.1f;

    [Tooltip("We're not allowing to flip our facing more often than this. Prevents spazzing out.")]
    [SerializeField]
    protected float minimumDurationBetweenFlips = 0.3f;

    [Tooltip("Set this bool in our Animator to set whether or not we're moving.")]
    [SerializeField]
    protected string animatorBoolMoving = "moving";

    [Tooltip("Set this bool in our Animator to set whether or not we're dead.")]
    [SerializeField]
    protected string animatorBoolDead = "dead";

    [Tooltip("Play these sounds as we walk.")]
    [SerializeField]
    protected AudioClip[] footstepSounds;

    [Tooltip("Wait this long between playing footstep sounds.")]
    [SerializeField]
    protected float timeBetweenFootstepSounds = 0.4f;

    [Tooltip("Use this audio source for footstep sounds.")]
    [SerializeField]
    protected AudioSource footstepAudioSource;

    [Tooltip("Play these sounds when hitting a solid wall.")]
    [SerializeField]
    protected AudioClip[] wallBumpSounds;

    [Tooltip("Wait this long between playing wall bump sounds.")]
    [SerializeField]
    protected float timeBetweenWallBumpSounds = 0.1f;

    protected float lastFlippedTime;
    protected float lastFootstepSoundTime;
    protected int lastPlayedFootstepSoundIndex;
    protected float lastWallBumpSoundTime;

    protected AudioSource audioSource;

    public virtual void Awake()
    {
        audioSource = GetComponent<AudioSource>();
    }

    public virtual void RefreshMovement( Vector2 velocity )
    {
        bool isMoving = velocity.sqrMagnitude >= walkingSpeedThreshold * walkingSpeedThreshold;
        IsMoving = isMoving;

        if ( isMoving )
        {
            IsFacingRight = velocity.x >= 0;
        }
    }

    protected bool _isMoving;
    public virtual bool IsMoving
    {
        get
        {
            return _isMoving;
        }
        set
        {
            _isMoving = value;
            animator.SetBool( animatorBoolMoving, value );
        }
    }

    public virtual void Update()
    {
        if ( IsMoving )
        {
            PlayFootstepSound();
        }
    }

    public virtual bool IsFacingRight
    {
        set
        {
            if ( Time.time - lastFlippedTime >= minimumDurationBetweenFlips )
            {
                Vector3 scale = transform.localScale;
                scale.x = Mathf.Abs( scale.x ) * ( value ? 1 : -1 );
                transform.localScale = scale;

                lastFlippedTime = Time.time;
            }
        }
    }

    public void PlayFootstepSound()
    {
        if ( Time.time - lastFootstepSoundTime < timeBetweenFootstepSounds || footstepSounds.Length <= 0 )
        {
            return;
        }

        lastFootstepSoundTime = Time.time;
        lastPlayedFootstepSoundIndex = ( lastPlayedFootstepSoundIndex + 1 ) % footstepSounds.Length;
        Sounds.Play( footstepSounds[lastPlayedFootstepSoundIndex], footstepAudioSource );
    }

    public void BumpedWall()
    {
        if ( Time.time - lastWallBumpSoundTime >= timeBetweenWallBumpSounds )
        {
            Sounds.Play( wallBumpSounds );
            lastWallBumpSoundTime = Time.time;
        }
    }

    protected bool _isDead;
    public virtual bool IsDead
    {
        get
        {
            return _isDead;
        }
        set
        {
            _isDead = value;
            animator.SetBool( animatorBoolDead, value );
        }
    }

    //TODO - any additional animations to be supported
}
