﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public class PatrollingEnemyController : EnemyController
{
    [Tooltip("The view for the player")]
    [SerializeField]
    protected CharacterView view;

    [Tooltip("The list of transforms this enemy will patrol between")]
    [SerializeField]
    protected List<Transform> patrolPoints;

    [Tooltip("The cutoff for when the character reaches the patrol point")]
    [SerializeField]
    protected float reachedPointCutoff;

    [Tooltip("The cutoff for when the character gets attracted to the attract rune")]
    [SerializeField]
    protected float attractCutoff;


    private int currentPatrolIndex;

    // Start is called before the first frame update
    void Start()
    {
        currentPatrolIndex = 0;
    }

    // Update is called once per frame
    void Update()
    {
        if (patrolPoints.Count <= 0)
        {
            return;
        }


        view.RefreshMovement(mover.Velocity);

        Collider2D attractionCollider = base.FindClosestTargetInRange(transform.position, attractionRadius, attractionLayerMask);

        if (attractionCollider != null)
        {
            Vector2 closestPointToPatrolPath = GetClosestPointFromListOfPoints(patrolPoints, transform.position);

            if( Vector2.Distance(closestPointToPatrolPath, transform.position) < attractCutoff)
            {
                base.MoveTowardsPoint(attractionCollider.transform.position);
                return;
            }
        }
      
        if (HasReachedCurrentPoint())
        {
            currentPatrolIndex++;

            if (currentPatrolIndex >= patrolPoints.Count)
            {
                currentPatrolIndex = 0;
          }
        }

        base.MoveTowardsPoint(patrolPoints[currentPatrolIndex].position);
    }

    private bool HasReachedCurrentPoint()
    {
        float currentDistance = Vector2.Distance(patrolPoints[currentPatrolIndex].position, transform.position);
        if (currentDistance < reachedPointCutoff)
        {
            return true;
        }
        return false;
    }

    public Vector2 GetClosestPointFromListOfPoints( List<Transform> points, Vector2 pointFrom)
    {
        float closestDistance = float.MaxValue;
        Vector2 closestPoint = pointFrom;

        for( int i = 0; i < points.Count; i++ )
        {
            if( points.Count > i+1 )
            {
                Vector2 closestPointThisLine = GetClosestPointOnLineSegment( points[i].position, points[i+1].position, pointFrom);

                float distanceFrom = Vector2.Distance(closestPointThisLine, pointFrom);

                if (distanceFrom < closestDistance)
                {
                    closestPoint = closestPointThisLine;
                    closestDistance = distanceFrom;
                }
            }
        }

        return closestPoint;
    }

    public Vector2 GetClosestPointOnLineSegment(Vector2 lineStartPoint, Vector2 lineEndPoint, Vector2 pointFrom)
    {
        Vector2 AP = pointFrom - lineStartPoint;       //Vector from A to P   
        Vector2 AB = lineEndPoint - lineStartPoint;       //Vector from A to B  

        float magnitudeAB = AB.SqrMagnitude();     //Magnitude of AB vector (it's length squared)     
        float ABAPproduct = Vector2.Dot(AP, AB);    //The DOT product of a_to_p and a_to_b     
        float distance = ABAPproduct / magnitudeAB; //The normalized "distance" from a to your closest point  

        if (distance < 0)     //Check if P projection is over vectorAB     
        {
            return lineStartPoint;

        }
        else if (distance > 1)
        {
            return lineEndPoint;
        }
        else
        {
            return lineStartPoint + AB * distance;
        }
    }

    public void OnDrawGizmos()
    {
        Gizmos.color = Color.red;
        Gizmos.DrawWireSphere(transform.position, attractionRadius);
    }
}
