﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

[RequireComponent(typeof(Mover))]
[RequireComponent(typeof(Destructible))]
[RequireComponent(typeof(CircleCollider2D))]
//This controls the Knight's movement. It follows a simple layer-style architecture for AI.
// 1) Stop if a stopping layer mask has been hit.
// 2) Attract to the player dropped attractor.
// 3) Repel from the player dropped repeller.
// 4) Avoid the avoided layer mask. (handled with a simple steering algorithm)
public class KnightController : MoverModifier
{
    [Tooltip("We need a view.")]
    [SerializeField]
    protected KnightView view;

    [Tooltip("He's attracted to everything within this LayerMask.")]
    [SerializeField]
    protected LayerMask attractionLayerMask;

    [Tooltip("If he's this close to an attractor, he'll move towards it.")]
    [SerializeField]
    protected float attractionRadius = 5.0f;

    [Tooltip("How quickly he turns towards what he's attracted.")]
    [SerializeField]
    protected float attractionSpeed = 1.0f;

    [Tooltip("He's repulsed from everything within this LayerMask.")]
    [SerializeField]
    protected LayerMask repulsionLayerMask;

    [Tooltip("If he's this close to a repeller, he'll run away from it.")]
    [SerializeField]
    protected float repulsionRadius = 5.0f;

    [Tooltip("How quickly he turns away from what he's repelled by.")]
    [SerializeField]
    protected float repulsionSpeed = 1.0f;

    [Tooltip("If a raycast to a trarget hits any of this, then he will not be attracted / repelled.")]
    [SerializeField]
    protected LayerMask blockAttractionAndRepulsionThroughLayerMask;

    [Tooltip("Considers 2D colliders in this LayerMask to be something he can't walk through (but he won't avoid it, he'll just stop).")]
    [SerializeField]
    protected LayerMask stoppingLayerMask;

    [Tooltip("Will steer away from 2D colliders in this LayerMask.")]
    [SerializeField]
    protected LayerMask avoidedLayerMask;

    [Tooltip("Can see this far ahead of him in terms of avoidance.")]
    [SerializeField]
    protected float avoidedRadius = 2.0f;

    [Tooltip("Looks along this cone of vision in terms of avoidance.")]
    [SerializeField]
    protected float coneOfVisionInDegrees = 60.0f;

    [Tooltip("How quickly we turn away from avoided obstacles.")]
    [SerializeField]
    protected float avoidedSpeed = 3.0f;

    [Tooltip("The more of these, the more raycasts we do along our cone ahead. More precise, less performant.")]
    [SerializeField]
    protected int raycastCount = 16;

    [Tooltip("Delay before we tell the Game that we've actually won or lost. Logic should be moved elsewhere.")]
    [SerializeField]
    protected float winLoseDelay = 3.0f;

    protected Mover mover;
    protected Destructible destructible;
    protected CircleCollider2D circleCollider;

    protected Collider2D[] circleCastResults = new Collider2D[64];
    protected RaycastHit2D[] raycastResults = new RaycastHit2D[64];

    protected bool isStopped;

    public virtual void Awake()
    {
        mover = GetComponent<Mover>();
        destructible = GetComponent<Destructible>();
        circleCollider = GetComponent<CircleCollider2D>();
    }

    public virtual void Start()
    {
        //hacky hacky joy joy... are we in the first level?
        if ( SceneLoader.GetLoadedSceneName().Substring( "Level".Length ) == "0" )
        {
            view.PlayIntroEmotes();
        }
    }

    public virtual void Update()
    {
        view.RefreshMovement( isStopped ? Vector2.zero : mover.Velocity );

        view.IsDead = destructible.IsDying;

        if ( destructible.IsDying )
        {
            Lose();
        }
    }

    public override Vector2 ModifyVelocity( Vector2 vel )
    {
        if ( destructible.IsDying )
        {
            return Vector2.zero;
        }

        bool shouldStop = ShouldStop( vel * Time.deltaTime );
        view.IsStopped = shouldStop;
        if ( shouldStop )
        {
            view.IsAttracting = false;
            view.IsRepelling = false;
            //with this commented out, he can accrue velocity changes even while stopped.
            //in order to maintain his velocity before beind stopped, uncomment.
            //return vel;
        }

        vel = DoAvoidance( vel );

        Collider2D repulsionTarget = FindClosestRepulsionTargetInRange();
        view.IsRepelling = !shouldStop && repulsionTarget != null;
        if (repulsionTarget != null)
        {
            view.IsAttracting = false;
            return DoRepulsion( vel, repulsionTarget, !shouldStop );
        }

        Collider2D attractionTarget = FindClosestAttractionTargetInRange();
        view.IsAttracting = !shouldStop && attractionTarget != null;
        if ( attractionTarget != null )
        {
            return DoAttraction( vel, attractionTarget, !shouldStop );
        }

        //return DoAvoidance( vel );
        return vel;
    }

    public override Vector2 ModifyFrameDelta( Vector2 frameDelta )
    {
        if ( destructible.IsDying )
        {
            return Vector2.zero;
        }
        
        isStopped = ShouldStop( frameDelta );
        if ( isStopped )
        {
            return DoStopping( frameDelta, true );
        }

        return frameDelta;
    }

    public void PickUpStar()
    {
        view.PickUpStar();
    }

    protected bool _isWinning;
    public void Win()
    {
        if ( _isWinning )
        {
            return;
        }

        destructible.IsInvulnerable = true;
        view.HasWon = true;
        _isWinning = true;
        Invoke( "ActuallyWin", winLoseDelay );
    }

    protected void ActuallyWin()
    {
        //nothing
    }

    protected bool _isLosing;
    public void Lose()
    {
        if ( _isLosing )
        {
            return;
        }

        _isLosing = true;
        World.Instance.CurrentLevel.Lose();
        Invoke( "ActuallyLose", winLoseDelay );
    }

    protected void ActuallyLose()
    {
        //nothing
    }

    protected virtual Collider2D FindClosestAttractionTargetInRange()
    {
        return FindClosestTargetInRange( attractionRadius, attractionLayerMask );
    }

    protected virtual Collider2D FindClosestRepulsionTargetInRange()
    {
        return FindClosestTargetInRange( repulsionRadius, repulsionLayerMask );
    }

    protected virtual bool ShouldStop( Vector2 frameDelta )
    {
        //TODO wouldn't be surprised if this has issues with any sort of moving obstacle.
        //once he gets stuck on an obstacle, he'll never be free. This also doesn't
        //allow smooth stops unless we increase the radius or something.
        //the better solution is logic to figure out how to get away from the obstacle...
        //but the steering algorithm does that, theoretically?
        Vector2 targetPos = mover.Position;// + frameDelta; //removed frameDelta to just use his current position
        int hitCount = Physics2D.OverlapCircleNonAlloc( targetPos, circleCollider.radius, circleCastResults, stoppingLayerMask );
        return hitCount > 0;
    }

    protected virtual Vector2 DoAttraction( Vector2 vel, Collider2D target, bool allowEmotes )
    {
        if ( allowEmotes )
        {
            view.PlayEmote( view.Emotes.attracted );
        }

        return GetMovementTowardsTarget( vel, mover.Position, (Vector2) target.transform.position, attractionSpeed );
    }

    protected virtual Vector2 DoRepulsion( Vector2 vel, Collider2D target, bool allowEmotes )
    {
        if ( allowEmotes )
        {
            view.PlayEmote( view.Emotes.repulsed );
        }

        return GetMovementTowardsTarget( vel, (Vector2) target.transform.position, mover.Position, repulsionSpeed );
    }

    protected virtual Vector2 DoStopping( Vector2 vel, bool allowEmotes )
    {
        if ( allowEmotes )
        {
            view.PlayEmote( view.Emotes.stopped );
        }

        return Vector2.zero;
    }

    protected virtual Vector2 DoAvoidance( Vector2 vel )
    {
        float leftWeight = 0;
        float rightWeight = 0;

        float facingAngleRad = mover.FacingAngleRadians;
        for ( float angleDeltaDeg = -coneOfVisionInDegrees / 2.0f;
                    angleDeltaDeg <= coneOfVisionInDegrees / 2.0f; 
                    angleDeltaDeg += coneOfVisionInDegrees / raycastCount )
        {
            //no reason to bother looking straight down the middle for the sake of steering
            if ( !Mathf.Approximately( angleDeltaDeg, 0.0f ) )
            {
                //see if there's anything ahead
                float angleRad = angleDeltaDeg * Mathf.Deg2Rad + facingAngleRad;
                Vector2 dir = new Vector2( Mathf.Cos( angleRad ), Mathf.Sin( angleRad ) );
                int hitCount = Physics2D.RaycastNonAlloc( mover.Position, dir, raycastResults, avoidedRadius, avoidedLayerMask );
                if ( hitCount > 0 )
                {
                    //count whichever side we saw the thing on
                    if ( angleDeltaDeg < 0 )
                    {
                        leftWeight += avoidedRadius - raycastResults[0].distance;
                    }
                    else if ( angleDeltaDeg > 0 )
                    {
                        rightWeight += avoidedRadius - raycastResults[0].distance;
                    }
                }

                Debug.DrawLine( mover.Position, mover.Position + dir * avoidedRadius, hitCount > 0 ? Color.red : Color.green );
            }
        }

        //if we didn't see anything at all, then carry on
        if ( Mathf.Approximately( leftWeight, 0 ) && Mathf.Approximately( rightWeight, 0 ) )
        {
            return vel;
        }

        Vector2 rightVel = new Vector2( vel.y, -vel.x ).normalized;

        //if we saw more on our left, turn in that direction
        if ( leftWeight > rightWeight )
        {
            return GetMovementTowardsTarget( vel, mover.Position, mover.Position - rightVel, avoidedSpeed );
        }

        //otherwise turn right
        return GetMovementTowardsTarget( vel, mover.Position, mover.Position + rightVel, avoidedSpeed );
    }

    protected virtual Collider2D FindClosestTargetInRange( float radius, LayerMask layerMask )
    {
        int hitCount = Physics2D.OverlapCircleNonAlloc( mover.Position, radius, circleCastResults, layerMask );

        Collider2D closestCollider = null;
        float closestSqrDistance = 0.0f;

        for ( int resultIndex = 0; resultIndex < hitCount; resultIndex++ )
        {
            Collider2D coll = circleCastResults[resultIndex];
            Vector2 diff = ( (Vector2) coll.transform.position - mover.Position );
            float sqrDistance = diff.sqrMagnitude;
            if ( closestCollider == null || sqrDistance < closestSqrDistance )
            {
                if ( !Physics2D.Raycast( mover.Position, diff.normalized, Mathf.Sqrt( sqrDistance ), blockAttractionAndRepulsionThroughLayerMask ) )
                {
                    closestSqrDistance = sqrDistance;
                    closestCollider = coll;
                }
            }
        }

        return closestCollider;
    }

    protected virtual Vector2 GetMovementTowardsTarget( Vector2 vel, Vector2 originPos, Vector2 targetPos, float speed )
    {
        float velMag = mover.InitialSpeed; //vel.magnitude;
        Vector2 dir = ( targetPos - originPos ).normalized;
        vel = Vector3.ClampMagnitude( vel + dir * speed * Time.deltaTime, velMag );

        return vel;
    }
}
