﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public class KnightView : CharacterView
{
    [Tooltip("This is our voice bubble, which we can use to say stuff.")]
    [SerializeField]
    protected VoiceBubbleView voiceBubbleView;

    [Tooltip("These are the different things he says, both audio and string in the balloon.")]
    [SerializeField]
    protected EmoteList emotes;

    [Tooltip("How long between intro emotes.")]
    [SerializeField]
    protected float timeBetweenIntroEmotes = 1.0f;

    [Tooltip("The minimum time between playing other emotes.")]
    [SerializeField]
    protected float minimumTimeBetweenEmotes = 10.0f;

    [Tooltip("Set this bool in our Animator to set whether or not we've won.")]
    protected string animatorBoolWon = "won";

    [Tooltip("Set this bool in our Animator to set whether or not we're being attracted.")]
    protected string animatorBoolAttracting = "attracting";

    [Tooltip("Set this bool in our Animator to set whether or not we're being repelled.")]
    protected string animatorBoolRepelling = "repelling";

    [Tooltip("Set this bool in our Animator to set whether or not we're being stopped.")]
    protected string animatorBoolStopped = "stopped";

    public EmoteList Emotes
    {
        get
        {
            return emotes;
        }
    }

    [System.Serializable]
    public struct EmoteList
    {
        public Emote introSquire;
        public Emote introFindWoman;
        public Emote introLeadMe;
        public Emote introUseRune;
        public Emote[] attracted;
        public Emote[] repulsed;
        public Emote[] stopped;
        public Emote[] wonLevel;
        public Emote[] died;
        public Emote[] gotStar;
    }

    [System.Serializable]
    public struct Emote
    {
        public string text;
        public AudioClip audio;
    }

    public enum IntroEmoteStep
    {
        Unstarted,
        Squire,
        FindWoman,
        LeadMe,
        UseRune,
        Complete
    };

    protected IntroEmoteStep introEmoteStep;
    protected static Dictionary<string,int> emotePlayCounts;

    protected float lastTimePlayedEmote;

    public override bool IsFacingRight
    {
        set
        {
            base.IsFacingRight = value;
            voiceBubbleView.TextIsFlipped = transform.localScale.x < 0;
        }
    }

    public override bool IsDead
    {
        set
        {
            if ( IsDead != value )
            {
                PlayEmote( emotes.died, true );
            }
            base.IsDead = value;
        }
    }

    public virtual bool IsAttracting
    {
        set
        {
            animator.SetBool( animatorBoolAttracting, value );
        }
    }

    public virtual bool IsRepelling
    {
        set
        {
            animator.SetBool( animatorBoolRepelling, value );
        }
    }

    public virtual bool IsStopped
    {
        set
        {
            animator.SetBool( animatorBoolStopped, value );
        }
    }

    protected bool _hasWon;
    public virtual bool HasWon
    {
        get
        {
            return _hasWon;
        }
        set
        {
            if ( value != _hasWon )
            {
                _hasWon = value;
                PlayEmote( emotes.wonLevel, true );
            }
            animator.SetBool( animatorBoolWon, value );
        }
    }

    public override void Awake()
    {
        base.Awake();

        if ( emotePlayCounts == null )
        {
            emotePlayCounts = new Dictionary<string, int>();
        }
    }

    public void PickUpStar()
    {
        PlayEmote( emotes.gotStar, true );
    }

    public void PlayIntroEmotes()
    {
        introEmoteStep = IntroEmoteStep.Unstarted;
        GoToNextIntroEmote();
    }

    public void PlayEmote( Emote emote, bool force )
    {
        if ( !force && Time.time - lastTimePlayedEmote < minimumTimeBetweenEmotes )
        {
            return;
        }

        voiceBubbleView.SetText( emote.text );
        Sounds.Play( emote.audio, audioSource );

        emotePlayCounts[emote.text] = GetEmotePlayCount( emote ) + 1;
        lastTimePlayedEmote = Time.time;
    }

    public void PlayEmote( Emote emote )
    {
        PlayEmote( emote, false );
    }

    public void PlayEmote( Emote[] emotes )
    {
        PlayEmote( emotes, false );
    }

    public void PlayEmote( Emote[] emotes, bool force )
    {
        if ( !force && Time.time - lastTimePlayedEmote < minimumTimeBetweenEmotes )
        {
            return;
        }

        if ( emotes.Length <= 0 )
        {
            return;
        }
        else if ( emotes.Length <= 1 )
        {
            PlayEmote( emotes[0] );
        }

        //I'm sure we can make this more efficient
        int lowestPlayCount = GetEmotePlayCount( emotes[0] );
        foreach( Emote emote in emotes )
        {
            int playCount = GetEmotePlayCount( emote );
            if ( playCount < lowestPlayCount )
            {
                lowestPlayCount = playCount;
            }
        }

        List<Emote> choices = new List<Emote>();
        foreach( Emote emote in emotes )
        {
            if ( GetEmotePlayCount( emote ) == lowestPlayCount )
            {
                choices.Add( emote );
            }
        }

        PlayEmote( choices[ Random.Range( 0, choices.Count ) ], force );
    }

    protected int GetEmotePlayCount( Emote emote )
    {
        return emotePlayCounts.ContainsKey( emote.text ) ? emotePlayCounts[ emote.text ] : 0;
    }

    protected void GoToNextIntroEmote()
    {
        if ( introEmoteStep >= IntroEmoteStep.Complete )
        {
            return;
        }

        if ( HasWon )
        {
            introEmoteStep = IntroEmoteStep.Complete;
            return;
        }
        
        introEmoteStep++;

        if ( introEmoteStep >= IntroEmoteStep.Complete )
        {
            return;
        }

        Emote emote = emotes.introSquire;
        switch ( introEmoteStep )
        {
            case IntroEmoteStep.Squire:
                break;
            case IntroEmoteStep.FindWoman:
                emote = emotes.introFindWoman;
                break;
            case IntroEmoteStep.LeadMe:
                emote = emotes.introLeadMe;
                break;
            case IntroEmoteStep.UseRune:
                emote = emotes.introUseRune;
                break;
        }

        PlayEmote( emote, true );

        Invoke( "GoToNextIntroEmote", timeBetweenIntroEmotes );
    }
}
