﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

[RequireComponent(typeof(Mover))]
[RequireComponent(typeof(FrictionMoverModifier))]
[RequireComponent(typeof(InputController))]
public class PlayerController : MonoBehaviour
{
    [Tooltip("The view for the player")]
    [SerializeField]
    protected PlayerView view;

    [Tooltip("The speed at which the player speeds up")]
    [SerializeField]
    protected float walkSpeed;

    [Tooltip("The maximum speed the player can move")]
    [SerializeField]
    protected float maxSpeed;

    [Tooltip("The LayerMask for attraction objects player can pick up")]
    [SerializeField]
    protected LayerMask attractionLayerMask;

    [Tooltip("The LayerMask for repulsion objects player can pick up")]
    [SerializeField]
    protected LayerMask repulsionLayerMask;

    [Tooltip("The LayerMask for stop objects player can pick up")]
    [SerializeField]
    protected LayerMask stopLayerMask;

    [Tooltip("How close you have to be to pickup items")]
    [SerializeField]
    protected float pickupRadius = 1f;

    protected Mover mover;
    protected FrictionMoverModifier moverModifier;
    protected InputController inputController;
    protected Collider2D[] circleCastResults = new Collider2D[64];
    protected Rune heldRune;

    protected void Awake()
    {
        mover = GetComponent<Mover>();
        moverModifier = GetComponent<FrictionMoverModifier>();
        inputController = GetComponent<InputController>();

        inputController.OnDirectionHeld = OnDirectionHeld;
        inputController.OnDirectionUp = OnDirectionUp;
        inputController.OnInteractDown = OnInteractDown;
        inputController.OnInteractUp = OnInteractUp;

        //horrible hack
        Camera.main.GetComponent<AudioListener>().enabled = false;
    }

    protected void Update()
    {
        view.RefreshMovement(mover.Velocity);
    }

    protected bool _isWinning;
    public void Win()
    {
        if (_isWinning)
        {
            return;
        }

        mover.SetVelocity(new Vector2(0f,0f));
        inputController.AllowInput = false;
        view.HasWon = true;
        _isWinning = true;
    }

    public void Lose()
    {
        mover.SetVelocity(new Vector2(0f, 0f));
        inputController.AllowInput = false;
    }

    public void OnCollisionEnter2D( Collision2D col )
    {
        view.BumpedWall();
    }

    protected void OnDirectionHeld(Vector2 direction)
    {
        ApplyVelocity( direction, walkSpeed );
        moverModifier.Enabled = false;
    }

    protected void OnDirectionUp( Vector2 direction )
    {
        moverModifier.Enabled = true;
    }

    protected void OnInteractDown()
    {
        Collider2D interactedCollider = FindClosestAttractionTargetInRange();
        if(interactedCollider == null )
        {
            interactedCollider = FindClosestRepulsionTargetInRange();
        }
        if (interactedCollider == null)
        {
            interactedCollider = FindClosestStopTargetInRange();
        }
        if (interactedCollider == null)
        {
            return;
        }

        heldRune = interactedCollider.GetComponent<Rune>();

        view.SetHeldSprite(heldRune.MainSpriteRenderer.sprite);
        view.HasRune = true;

        heldRune.gameObject.SetActive(false);
    }

    protected void OnInteractUp()
    {
        if(heldRune == null)
        {
            return;
        }

        heldRune.gameObject.SetActive(true);
        heldRune.transform.position = transform.position;

        view.SetHeldSprite(null);
        view.HasRune = false;
        heldRune = null;
    }

    protected void ApplyVelocity( Vector2 direction, float speed )
    {
        float velMag = mover.InitialSpeed; //vel.magnitude;
        Vector2 velocity = Vector3.ClampMagnitude(mover.Velocity + direction * speed * Time.deltaTime, maxSpeed);

        mover.SetVelocity(velocity);
    }

    protected virtual Collider2D FindClosestAttractionTargetInRange()
    {
        return FindClosestTargetInRange(pickupRadius, attractionLayerMask);
    }

    protected virtual Collider2D FindClosestRepulsionTargetInRange()
    {
        return FindClosestTargetInRange(pickupRadius, repulsionLayerMask);
    }

    protected virtual Collider2D FindClosestStopTargetInRange()
    {
        return FindClosestTargetInRange(pickupRadius, stopLayerMask);
    }

    protected virtual Collider2D FindClosestTargetInRange(float radius, LayerMask layerMask)
    {
        int hitCount = Physics2D.OverlapCircleNonAlloc(mover.Position, radius, circleCastResults, layerMask);

        Collider2D closestCollider = null;
        float closestSqrDistance = 0.0f;

        for (int resultIndex = 0; resultIndex < hitCount; resultIndex++)
        {
            Collider2D coll = circleCastResults[resultIndex];
            float sqrDistance = (mover.Position - (Vector2)coll.transform.position).sqrMagnitude;
            if (closestCollider == null || sqrDistance < closestSqrDistance)
            {
                closestSqrDistance = sqrDistance;
                closestCollider = coll;
            }
        }

        return closestCollider;
    }
}
