//the music gets annoying when editing and listening to music...
#if UNITY_EDITOR
//#define NO_MUSIC
#endif

using UnityEngine;

public class MusicController : MonoBehaviour
{
    public const string MUSIC_VOLUME_PERCENT_KEY = "musicVolume";

    public float globalCrossFadeTime = 0.5f;
    
    protected float currentCrossFadeLength;
    protected float currentCrossFadeTime;
    
    protected bool fadingOut;
    protected bool fadingIn;
    protected AudioClip swapToClip;
    protected AudioSource audioSource;
    
    protected static MusicController instance;
    public static MusicController Instance
    {
        get
        {
            return instance;
        }
    }
    
    public void Awake()
    {
        if ( instance != null & instance != this )
        {
            Object.DestroyImmediate( gameObject );
            return;
        }
        
        instance = this;
        Object.DontDestroyOnLoad( gameObject );
        
        audioSource = GetComponent<AudioSource>();

        //no doppler effects for music
        audioSource.dopplerLevel = 0.0f;
        
        //no spatial blend either
        audioSource.spatialBlend = 0.0f;
    }
    
    public void Play( AudioClip music )
    {
        Play( music, globalCrossFadeTime );
    }
    
    public void Play( AudioClip music, float crossFadeTime )
    {
#if NO_MUSIC
        return;
#else
        audioSource.loop = true;
        RefreshVolume();
        
        //if already playing this music, do nothing
        if ( audioSource.isPlaying && audioSource.clip == music )
        {
            //edge case, fade back in if we were fading out
            if ( fadingOut )
            {
                fadingOut = false;
                fadingIn = true;
                currentCrossFadeTime = currentCrossFadeLength - currentCrossFadeTime;
            }
            
            return;
        }
        
        if ( !audioSource.isPlaying || crossFadeTime == 0.0f )
        {
            audioSource.Stop();
            audioSource.clip = music;
            audioSource.Play();
            FadeIn( crossFadeTime );
        }
        else
        {
            swapToClip = music;
            FadeOut( crossFadeTime );
        }
#endif
    }
    
    public void PlayOnce( AudioClip music )
    {
        PlayOnce( music, globalCrossFadeTime );
    }
    
    public void PlayOnce( AudioClip music, float crossFadeTime )
    {
        Play( music, crossFadeTime );
        audioSource.loop = false;
    }
    
    public void Stop()
    {
        Stop( globalCrossFadeTime );
    }
    
    public void Stop( float fadeOutTime )
    {
        FadeOut( fadeOutTime );
    }
    
    public void FadeIn( float time )
    {
        if ( time > 0.0f )
        {
            currentCrossFadeLength = time;
            currentCrossFadeTime = 0.0f;
            SetVolume( 0.0f );
            fadingIn = true;
            fadingOut = false;
        }
        else
        {
            SetVolume( 1.0f );
            fadingOut = false;
            fadingIn = false;
        }
    }
    
    public void FadeOut( float time )
    {
        if ( time > 0.0f )
        {
            currentCrossFadeLength = time;
            currentCrossFadeTime = 0.0f;
            SetVolume( 1.0f );
            fadingIn = false;
            fadingOut = true;
        }
        else
        {
            currentCrossFadeLength = 0.0f;
            FinishedFadingOut();
        }
    }
    
    public void Update()
    {
        if ( fadingOut )
        {
            currentCrossFadeTime += Time.deltaTime;
            SetVolume( Mathf.Clamp( 1.0f - currentCrossFadeTime / currentCrossFadeLength, 0.0f, 1.0f ) );
            
            if ( currentCrossFadeTime >= currentCrossFadeLength )
            {
                FinishedFadingOut();
            }
        }
        else if ( fadingIn )
        {
            currentCrossFadeTime += Time.deltaTime;
            SetVolume( Mathf.Clamp( currentCrossFadeTime / currentCrossFadeLength, 0.0f, 1.0f ) );
            
            if ( currentCrossFadeTime >= currentCrossFadeLength )
            {
                fadingIn = false;
            }
        }
    }

    public void RefreshVolume()
    {
        SetVolume( 1.0f );
    }
    
    protected void FinishedFadingOut()
    {
        SetVolume( 0.0f );
        fadingOut = false;
        audioSource.Stop();
        
        if ( swapToClip != null )
        {
            audioSource.clip = swapToClip;
            swapToClip = null;
            audioSource.Play();
            FadeIn( currentCrossFadeLength );
        }
    }

    public void SetVolume( float level )
    {
        audioSource.volume = level * PlayerPrefs.GetFloat( MUSIC_VOLUME_PERCENT_KEY, 100.0f ) / 100.0f;
    }
}