using UnityEngine;
using UnityEngine.SceneManagement;
using System.Collections.Generic;

//singleton that handles loading scenes in a nicer way. can automatically load scenes with a different
//suffix for landscape or portrait views, and will only load those scenes if they exist.
//relies on Globals to determine if the game is in landscape or not.
//to use, call the static function LoadScene(), and that will use the SceneLoader if it exists,
//or will just load the scene immediately if not (useful when launching weird scenes from the editor).
//also can have a SceneLoadingView attached (optional), which can respond to changes to the
//SceneLoadingModel in order to render overlays or loading bars, etc.
//you can also add listeners to get notified when loading has finished fading in or out.
public class SceneLoader : MonoBehaviour
{
    [Tooltip("How long the fading in and fading out phases last. informs the view.")]
    public float transitionTime = 0.5f;

    [Tooltip("Animator that we can use to animate transitions.")]
    [SerializeField]
    public Animator animator;

    [Tooltip("Boolean param for the animator to mark whether or not it's loading.")]
    [SerializeField]
    public string animatorBoolLoading = "loading";

    protected List<Listener> listeners;
    protected string levelToLoad;

    protected bool isLoading;
        
    protected static SceneLoader instance;
    public static SceneLoader Instance
    {
        get
        {
            return instance;
        }
    }
    
    public void Awake()
    {
        if ( instance != null && instance != this )
        {
            Object.DestroyImmediate( gameObject );
            return;
        }
                
        instance = this;
        
        DontDestroyOnLoad( gameObject );
        
        if ( listeners == null )
        {
            listeners = new List<Listener>();
        }
    }
    
    //for convenience, you can call this even if there is no SceneLoader.
    //if you do, then it just straight up loads that scene, using the default of -landscape added on scene names.
    public static void LoadScene( string sceneName )
    {
        if ( Instance == null )
        {
            SceneManager.LoadScene( sceneName );
        }
        else
        {
            Instance.LoadLevel( sceneName );
        }
    }
    
    //for convenience, you can call this even if there is no SceneLoader.
    //if you do, then it removes a default of -landscape from the end of scene names.
    public static string GetLoadedSceneName()
    {
        string name = SceneManager.GetActiveScene().name;
        
        return name;
    }
    
    //for convenience, checks the currently loaded scene name against the passed scene name.
    public static bool SceneIsLoaded( string name )
    {
        return GetLoadedSceneName().Equals( name );
    }

    public string GetTargetSceneName()
    {
        return levelToLoad;
    }

    public bool IsLoadingScene( string levelName )
    {
        return levelToLoad == levelName; 
    }
    
    //are we in the midst of loading?
    public static bool IsLoading()
    {
        if ( Instance == null )
        {
            return false;
        }
        
        return Instance.isLoading;
    }
    
    public void AddListener( Listener newListener )
    {
        //hack for Awake() happening in indeterminte order for objects
        if ( listeners == null )
        {
            Awake();
        }

        if ( !listeners.Contains( newListener ) )
        {
            listeners.Add( newListener );
        }
    }
    
    public void RemoveListener( Listener removeListener )
    {
        listeners.Remove( removeListener );
    }

    public void AnimatorCallbackOverlayAppeared()
    {
        SendListenersUpdate( true, false, false );

        SceneManager.LoadScene( levelToLoad );
        levelToLoad = "";

        Invoke( "LoadingCompleted", 0.1f );
    }

    public void LoadingCompleted()
    {
        animator.SetBool( animatorBoolLoading, false );

        SendListenersUpdate( false, true, false );
    }

    public void AnimatorCallbackOverlayDisappeared()
    {
        SendListenersUpdate( false, false, true );
    }
    
    protected void LoadLevel( string setLevelToLoad )
    {
        levelToLoad = setLevelToLoad;
        animator.SetBool( animatorBoolLoading, true );
    }
    
    protected virtual void SendListenersUpdate( bool sendFadedIn, bool sendCompleted, bool sendFadedOut )
    {
        for ( int listenerIndex = listeners.Count - 1; listenerIndex >= 0; listenerIndex-- )
        {
            Listener listener = listeners[ listenerIndex ];
            if ( listener == null || ( listener is MonoBehaviour && ( listener as MonoBehaviour == null ) )
                 || ( listener as MonoBehaviour != null && ( listener as MonoBehaviour ).gameObject == null ) )
            {
                listeners.RemoveAt( listenerIndex );
                continue;
            }
            
            if ( sendFadedIn )
            {
                listener.LoadingFadedOut();
            }

            if ( sendCompleted )
            {
                listener.LoadingCompleted();
            }
            
            if ( sendFadedOut )
            {
                listener.LoadingFadedIn();
            }
        }
    }
    
    public interface Listener
    {
        void LoadingFadedOut();
        void LoadingCompleted();
        void LoadingFadedIn();
    }
}