﻿using UnityEngine;
using System.Collections.Generic;

public class Utility
{
    public static Collider MouseRaycast( int layerMask )
    {
        Ray mouseRay = GetMouseRay();
        RaycastHit hit;
        if ( Physics.Raycast( mouseRay, out hit, Mathf.Infinity, layerMask ) )
        {
            return hit.collider;
        }
        return null;
    }
    
    public static Collider2D MouseRaycast2D( int layerMask )
    {
        Ray mouseRay = GetMouseRay();
        RaycastHit2D hit = Physics2D.GetRayIntersection( mouseRay, Mathf.Infinity, layerMask );
        return hit.collider;
    }
    
    public static Ray GetMouseRay()
    {
        return GetCameraRayFromScreenPosition( new Vector3( Input.mousePosition.x, Input.mousePosition.y, 0.0f ) );
    }
    
    public static Vector3 GetMousePosition()
    {
        return GetWorldPositionFromRay( Utility.GetMouseRay() );
    }

    public static Ray GetTouchRay( Touch touch )
    {
        return GetCameraRayFromScreenPosition( touch.position );
    }

    public static Vector3 GetTouchPosition( Touch touch )
    {
        return GetWorldPositionFromRay( Utility.GetTouchRay( touch ) );
    }

    private static Ray GetCameraRayFromScreenPosition( Vector3 pos )
    {
        return Camera.main.ScreenPointToRay( pos );
    }

    private static Vector3 GetWorldPositionFromRay( Ray ray )
    {
        return ray.GetPoint( Mathf.Abs( Camera.main.transform.position.z ) );
    }
    
    public static Vector3 GetEaseInOutPosition( float time, float duration, Vector3 start, Vector3 end )
    {
        return GetNormalizedEaseInOutTime( time, duration ) * ( end - start ) + start;
    }
    
    public static float GetEaseInOutPosition( float time, float duration, float start, float end )
    {
        return GetNormalizedEaseInOutTime( time, duration ) * ( end - start ) + start;
    }
    
    public static float GetEaseInOutTime( float time, float duration )
    {
        return GetNormalizedEaseInOutTime( time, duration ) * duration;
    }

    public static float GetEaseTime( float t, float duration, Ease ease )
    {
        return GetNormalizedEase( t / duration, ease ) * duration;
    }

    public static float GetNormalizedEase( float t, Ease ease )
    {
        switch ( ease )
        {
            case Ease.In:
                return GetNormalizedEaseInTime( t, 1.0f );
            case Ease.Out:
                return GetNormalizedEaseOutTime( t, 1.0f );
            case Ease.InOut:
                return GetNormalizedEaseInOutTime( t, 1.0f );
        }
        return t;
    }

    public static float GetNormalizedEaseInOutTime( float time, float duration )
    {
        if ( time >= duration )
        {
            return 1.0f;
        }
        
        float t = time / ( duration / 2.0f );
        
        //ease in / out
        if ( t < 1.0f )
        {
            //ease in
            return 0.5f * t * t;
        }
        else
        {
            //ease out
            t -= 1.0f;
            return -0.5f * ( t * ( t - 2.0f ) - 1.0f );
        }
    }
    
    public static float GetNormalizedEaseInTime( float time, float duration )
    {
        if ( time >= duration )
        {
            return 1.0f;
        }
        
        float t = time / duration;
        return t * t;
    }
    
    public static float GetNormalizedEaseOutTime( float time, float duration )
    {
        if ( time >= duration )
        {
            return 1.0f;
        }
        
        float t = time / duration;
        return -1.0f * t * ( t - 2.0f ) + 0.0f;
    }
    
    public static Vector3 GetNormalized2DDirection( Vector3 toVec, Vector3 fromVec )
    {
        return GetNormalized2DVector3( toVec - fromVec );
    }
    
    public static Vector3 GetNormalized2DVector3( Vector3 vec )
    {
        vec.z = 0.0f;
        return vec.normalized;
    }
    
    public static int GetSign( float val )
    {
        if ( val == 0.0f )
        {
            return 0;
        }
        
        return (int) ( Mathf.Abs( val ) / val );
    }
    
    public static void ShuffleList<T>( List<T> list )
    {
        for ( int i = 0; i < list.Count; i++ )
        {
            int randIndex = Random.Range( 0, list.Count );
            T temp = list[i];
            list[i] = list[randIndex];
            list[randIndex] = temp;
        }
    }

    public static Color GetAlphaColor( Color c, float alpha )
    {
        c.a = alpha;
        return c;
    }

    public static void SetAlpha( SpriteRenderer spriteRenderer, float alpha )
    {
        spriteRenderer.color = GetAlphaColor( spriteRenderer.color, alpha );
    }

    const float TWO_PI = 6.28318530718f;
    public static float GetNormalizedAngleRad( float angleRad )
    {
        return  ( ( angleRad % TWO_PI ) + TWO_PI ) % TWO_PI;
    }
    
    public struct RectTransformData
    {
        public Vector2 anchoredPosition;
        public Vector2 anchorMin;
        public Vector2 anchorMax;
        public Vector2 offsetMin;
        public Vector2 offsetMax;
        public Vector2 pivot;
        public Vector2 sizeDelta;
        
        public RectTransformData( RectTransform rect )
        {
            anchoredPosition = rect.anchoredPosition;
            anchorMin = rect.anchorMin;
            anchorMax = rect.anchorMax;
            offsetMin = rect.offsetMin;
            offsetMax = rect.offsetMax;
            pivot = rect.pivot;
            sizeDelta = rect.sizeDelta;
        }
        
        public bool Equals( RectTransformData data )
        {
            return
                anchoredPosition == data.anchoredPosition &&
                anchorMin == data.anchorMin &&
                anchorMax == data.anchorMax &&
                offsetMin == data.offsetMin &&
                offsetMax == data.offsetMax &&
                pivot == data.pivot &&
                sizeDelta == data.sizeDelta;
        }
        
        public bool Matches( RectTransform rect )
        {
            return Equals( new RectTransformData( rect ) );
        }
        
        public static void LerpTransform( RectTransform targetRect, RectTransformData startData, RectTransformData targetData, float t )
        {
            targetRect.anchoredPosition = Vector2.Lerp( startData.anchoredPosition, targetData.anchoredPosition, t );
            targetRect.anchorMin = Vector2.Lerp( startData.anchorMin, targetData.anchorMin, t );
            targetRect.anchorMax = Vector2.Lerp( startData.anchorMax, targetData.anchorMax, t );
            targetRect.offsetMin = Vector2.Lerp( startData.offsetMin, targetData.offsetMin, t );
            targetRect.offsetMax = Vector2.Lerp( startData.offsetMax, targetData.offsetMax, t );
            targetRect.pivot = Vector2.Lerp( startData.pivot, targetData.pivot, t );
            targetRect.sizeDelta = Vector2.Lerp( startData.sizeDelta, targetData.sizeDelta, t );
        }
        
        public override string ToString()
        {
            return "RectTransform:" +
                    " anchoredPosition: " + anchoredPosition +
                    " anchorMin: " + anchorMin +
                    " anchorMax: " + anchorMax +
                    " offsetMin: " + offsetMin +
                    " offsetMax: " + offsetMax +
                    " pivot: " + pivot +
                    " sizeDelta: " + sizeDelta;
        }
    }

    [System.Serializable]
    public enum Ease 
    {
        Linear,
        In, 
        Out, 
        InOut
    }
}